@@ -2,25 +2,44 @@ import asyncio
 import aiohttp
 import functools
 import http.cookies
-import time
 import ssl
 import socket
-import weakref
+import sys
+import traceback
 import warnings
 
+from collections import defaultdict
+from hashlib import md5, sha1, sha256
+from itertools import chain
+from math import ceil
+from types import MappingProxyType
+
 from . import hdrs
 from .client import ClientRequest
 from .errors import ServerDisconnectedError
 from .errors import HttpProxyError, ProxyConnectionError
 from .errors import ClientOSError, ClientTimeoutError
+from .errors import FingerprintMismatch
 from .helpers import BasicAuth
 
 
 __all__ = ('BaseConnector', 'TCPConnector', 'ProxyConnector', 'UnixConnector')
 
+PY_34 = sys.version_info >= (3, 4)
+PY_343 = sys.version_info >= (3, 4, 3)
+
+HASHFUNC_BY_DIGESTLEN = {
+    16: md5,
+    20: sha1,
+    32: sha256,
+}
+
 
 class Connection(object):
 
+    _source_traceback = None
+    _transport = None
+
     def __init__(self, connector, key, request, transport, protocol, loop):
         self._key = key
         self._connector = connector
@@ -30,7 +49,27 @@ class Connection(object):
         self._loop = loop
         self.reader = protocol.reader
         self.writer = protocol.writer
-        self._wr = weakref.ref(self, lambda wr, tr=self._transport: tr.close())
+
+        if loop.get_debug():
+            self._source_traceback = traceback.extract_stack(sys._getframe(1))
+
+    if PY_34:
+        def __del__(self):
+            if self._transport is not None:
+                if self._loop.is_closed():
+                    return
+
+                self._connector._release(
+                    self._key, self._request, self._transport, self._protocol,
+                    should_close=True)
+
+                warnings.warn("Unclosed connection {!r}".format(self),
+                              ResourceWarning)
+                context = {'client_connection': self,
+                           'message': 'Unclosed connection'}
+                if self._source_traceback is not None:
+                    context['source_traceback'] = self._source_traceback
+                self._loop.call_exception_handler(context)
 
     @property
     def loop(self):
@@ -42,14 +81,20 @@ class Connection(object):
                 self._key, self._request, self._transport, self._protocol,
                 should_close=True)
             self._transport = None
-            self._wr = None
 
     def release(self):
-        if self._transport:
+        if self._transport is not None:
             self._connector._release(
-                self._key, self._request, self._transport, self._protocol)
+                self._key, self._request, self._transport, self._protocol,
+                should_close=False)
             self._transport = None
-            self._wr = None
+
+    def detach(self):
+        self._transport = None
+
+    @property
+    def closed(self):
+        return self._transport is None
 
 
 class BaseConnector(object):
@@ -62,9 +107,21 @@ class BaseConnector(object):
     :param loop: Optional event loop.
     """
 
+    _closed = True  # prevent AttributeError in __del__ if ctor was failed
+    _source_traceback = None
+
     def __init__(self, *, conn_timeout=None, keepalive_timeout=30,
-                 share_cookies=False, force_close=False, loop=None):
+                 share_cookies=False, force_close=False, limit=None,
+                 loop=None):
+        if loop is None:
+            loop = asyncio.get_event_loop()
+
+        self._closed = False
+        if loop.get_debug():
+            self._source_traceback = traceback.extract_stack(sys._getframe(1))
+
         self._conns = {}
+        self._acquired = defaultdict(list)
         self._conn_timeout = conn_timeout
         self._keepalive_timeout = keepalive_timeout
         if share_cookies:
@@ -74,17 +131,48 @@ class BaseConnector(object):
         self._share_cookies = share_cookies
         self._cleanup_handle = None
         self._force_close = force_close
+        self._limit = limit
+        self._waiters = defaultdict(list)
 
-        if loop is None:
-            loop = asyncio.get_event_loop()
         self._loop = loop
         self._factory = functools.partial(
             aiohttp.StreamProtocol, loop=loop,
             disconnect_error=ServerDisconnectedError)
 
         self.cookies = http.cookies.SimpleCookie()
-        self._wr = weakref.ref(
-            self, lambda wr, f=self._do_close, conns=self._conns: f(conns))
+
+    if PY_34:
+        def __del__(self):
+            if self._closed:
+                return
+            if not self._conns:
+                return
+
+            self.close()
+
+            warnings.warn("Unclosed connector {!r}".format(self),
+                          ResourceWarning)
+            context = {'connector': self,
+                       'message': 'Unclosed connector'}
+            if self._source_traceback is not None:
+                context['source_traceback'] = self._source_traceback
+            self._loop.call_exception_handler(context)
+
+    @property
+    def force_close(self):
+        """Ultimately close connection on releasing if True."""
+        return self._force_close
+
+    @property
+    def limit(self):
+        """The limit for simultaneous connections to the same endpoint.
+
+        Endpoints are the same if they are have equal
+        (host, port, is_ssl) triple.
+
+        If limit is None the connector has no limit (default).
+        """
+        return self._limit
 
     def _cleanup(self):
         """Cleanup unused transports."""
@@ -92,66 +180,101 @@ class BaseConnector(object):
             self._cleanup_handle.cancel()
             self._cleanup_handle = None
 
-        now = time.time()
+        now = self._loop.time()
 
         connections = {}
+        timeout = self._keepalive_timeout
+
         for key, conns in self._conns.items():
             alive = []
             for transport, proto, t0 in conns:
                 if transport is not None:
                     if proto and not proto.is_connected():
                         transport = None
-                    elif (now - t0) > self._keepalive_timeout:
-                        transport.close()
-                        transport = None
+                    else:
+                        delta = t0 + self._keepalive_timeout - now
+                        if delta < 0:
+                            transport.close()
+                            transport = None
+                        elif delta < timeout:
+                            timeout = delta
 
-                if transport:
+                if transport is not None:
                     alive.append((transport, proto, t0))
             if alive:
                 connections[key] = alive
 
         if connections:
-            self._cleanup_handle = self._loop.call_later(
-                self._keepalive_timeout, self._cleanup)
+            self._cleanup_handle = self._loop.call_at(
+                ceil(now + timeout), self._cleanup)
 
         self._conns = connections
-        self._wr = weakref.ref(
-            self, lambda wr, f=self._do_close, conns=self._conns: f(conns))
 
     def _start_cleanup_task(self):
         if self._cleanup_handle is None:
-            self._cleanup_handle = self._loop.call_later(
-                self._keepalive_timeout, self._cleanup)
+            now = self._loop.time()
+            self._cleanup_handle = self._loop.call_at(
+                ceil(now + self._keepalive_timeout), self._cleanup)
 
     def close(self):
         """Close all opened transports."""
-        self._do_close(self._conns)
+        if self._closed:
+            return
+        self._closed = True
 
-    @staticmethod
-    def _do_close(conns):
-        for key, data in conns.items():
-            for transport, proto, t0 in data:
+        if not self._loop.is_closed():
+            for key, data in self._conns.items():
+                for transport, proto, t0 in data:
+                    transport.close()
+
+            for transport in chain(*self._acquired.values()):
                 transport.close()
 
-        conns.clear()
+            if self._cleanup_handle:
+                self._cleanup_handle.cancel()
+
+        self._conns.clear()
+        self._acquired.clear()
+        self._cleanup_handle = None
+
+    @property
+    def closed(self):
+        """Is connector closed.
+
+        A readonly property.
+        """
+        return self._closed
 
     def update_cookies(self, cookies):
-        """Update shared cookies."""
+        """Update shared cookies.
+
+        Deprecated, use ClientSession instead.
+        """
         if isinstance(cookies, dict):
             cookies = cookies.items()
 
         for name, value in cookies:
-            if isinstance(value, http.cookies.Morsel):
-                # use dict method because SimpleCookie class modifies value
-                dict.__setitem__(self.cookies, name, value)
-            else:
+            if PY_343:
                 self.cookies[name] = value
+            else:
+                if isinstance(value, http.cookies.Morsel):
+                    # use dict method because SimpleCookie class modifies value
+                    dict.__setitem__(self.cookies, name, value)
+                else:
+                    self.cookies[name] = value
 
     @asyncio.coroutine
     def connect(self, req):
         """Get from pool or create new connection."""
         key = (req.host, req.port, req.ssl)
 
+        # use short-circuit
+        if self._limit is not None:
+            while len(self._acquired[key]) >= self._limit:
+                fut = asyncio.Future(loop=self._loop)
+                self._waiters[key].append(fut)
+                yield from fut
+
         transport, proto = self._get(key)
         if transport is None:
             try:
@@ -168,14 +291,17 @@ class BaseConnector(object):
                 raise ClientOSError(
                     'Cannot connect to host %s:%s ssl:%s' % key) from exc
 
-        return Connection(self, key, req, transport, proto, self._loop)
+        self._acquired[key].append(transport)
+        conn = Connection(self, key, req, transport, proto, self._loop)
+        return conn
 
     def _get(self, key):
         conns = self._conns.get(key)
+        t1 = self._loop.time()
         while conns:
             transport, proto, t0 = conns.pop()
             if transport is not None and proto.is_connected():
-                if (time.time() - t0) > self._keepalive_timeout:
+                if t1 - t0 > self._keepalive_timeout:
                     transport.close()
                     transport = None
                 else:
@@ -184,6 +310,26 @@ class BaseConnector(object):
         return None, None
 
     def _release(self, key, req, transport, protocol, *, should_close=False):
+        if self._closed:
+            # acquired connection is already released on connector closing
+            return
+
+        acquired = self._acquired[key]
+        try:
+            acquired.remove(transport)
+        except ValueError:  # pragma: no cover
+            # this may be result of undetermenistic order of objects
+            # finalization due garbage collection.
+            pass
+        else:
+            if self._limit is not None and len(acquired) < self._limit:
+                waiters = self._waiters[key]
+                while waiters:
+                    waiter = waiters.pop(0)
+                    if not waiter.done():
+                        waiter.set_result(None)
+                        break
+
         resp = req.response
 
         if not should_close:
@@ -211,12 +357,13 @@ class BaseConnector(object):
             conns = self._conns.get(key)
             if conns is None:
                 conns = self._conns[key] = []
-            conns.append((transport, protocol, time.time()))
+            conns.append((transport, protocol, self._loop.time()))
             reader.unset_parser()
 
             self._start_cleanup_task()
 
-    def _create_connection(self, req, *args, **kwargs):
+    @asyncio.coroutine
+    def _create_connection(self, req):
         raise NotImplementedError()
 
 
@@ -228,23 +375,36 @@ class TCPConnector(BaseConnector):
     """TCP connector.
 
     :param bool verify_ssl: Set to True to check ssl certifications.
+    :param bytes fingerprint: Pass the binary md5, sha1, or sha256
+        digest of the expected certificate in DER format to verify
+        that the certificate the server presents matches. See also
+        https://en.wikipedia.org/wiki/Transport_Layer_Security#Certificate_pinning
     :param bool resolve: Set to True to do DNS lookup for host name.
     :param family: socket address family
     :param args: see :class:`BaseConnector`
     :param kwargs: see :class:`BaseConnector`
     """
 
-    def __init__(self, *args, verify_ssl=True,
+    def __init__(self, *, verify_ssl=True, fingerprint=None,
                  resolve=False, family=socket.AF_INET, ssl_context=None,
                  **kwargs):
+        super().__init__(**kwargs)
+
         if not verify_ssl and ssl_context is not None:
             raise ValueError(
                 "Either disable ssl certificate validation by "
                 "verify_ssl=False or specify ssl_context, not both.")
 
-        super().__init__(*args, **kwargs)
-
         self._verify_ssl = verify_ssl
+
+        if fingerprint:
+            digestlen = len(fingerprint)
+            hashfunc = HASHFUNC_BY_DIGESTLEN.get(digestlen)
+            if not hashfunc:
+                raise ValueError('fingerprint has invalid length')
+            self._hashfunc = hashfunc
+        self._fingerprint = fingerprint
+
         self._ssl_context = ssl_context
         self._family = family
         self._resolve = resolve
@@ -255,6 +415,11 @@ class TCPConnector(BaseConnector):
         """Do check for ssl certifications?"""
         return self._verify_ssl
 
+    @property
+    def fingerprint(self):
+        """Expected ssl certificate fingerprint."""
+        return self._fingerprint
+
     @property
     def ssl_context(self):
         """SSLContext instance for https requests.
@@ -271,7 +436,7 @@ class TCPConnector(BaseConnector):
             elif _SSH_HAS_CREATE_DEFAULT_CONTEXT:
                 # Python 3.4+
                 sslcontext = ssl.create_default_context()
-            else:  # pragma: no cover
+            else:
                 # Fallback for Python 3.3.
                 sslcontext = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
                 sslcontext.options |= ssl.OP_NO_SSLv2
@@ -295,7 +460,7 @@ class TCPConnector(BaseConnector):
     @property
     def resolved_hosts(self):
         """The dict of (host, port) -> (ipaddr, port) pairs."""
-        return dict(self._resolved_hosts)
+        return MappingProxyType(self._resolved_hosts)
 
     def clear_resolved_hosts(self, host=None, port=None):
         """Remove specified host/port or clear all resolve cache."""
@@ -329,7 +494,8 @@ class TCPConnector(BaseConnector):
             return [{'hostname': host, 'host': host, 'port': port,
                      'family': self._family, 'proto': 0, 'flags': 0}]
 
-    def _create_connection(self, req, **kwargs):
+    @asyncio.coroutine
+    def _create_connection(self, req):
         """Create connection.
 
         Has same keyword arguments as BaseEventLoop.create_connection.
@@ -340,20 +506,34 @@ class TCPConnector(BaseConnector):
             sslcontext = None
 
         hosts = yield from self._resolve_host(req.host, req.port)
+        exc = None
 
-        while hosts:
-            hinfo = hosts.pop()
+        for hinfo in hosts:
             try:
-                return (yield from self._loop.create_connection(
-                    self._factory, hinfo['host'], hinfo['port'],
+                host = hinfo['host']
+                port = hinfo['port']
+                conn = yield from self._loop.create_connection(
+                    self._factory, host, port,
                     ssl=sslcontext, family=hinfo['family'],
                     proto=hinfo['proto'], flags=hinfo['flags'],
-                    server_hostname=hinfo['hostname'] if sslcontext else None,
-                    **kwargs))
-            except OSError as exc:
-                if not hosts:
-                    raise ClientOSError('Can not connect to %s:%s' %
-                                        (req.host, req.port)) from exc
+                    server_hostname=hinfo['hostname'] if sslcontext else None)
+                transport = conn[0]
+                has_cert = transport.get_extra_info('sslcontext')
+                if has_cert and self._fingerprint:
+                    sock = transport.get_extra_info('socket')
+                    # gives DER-encoded cert as a sequence of bytes (or None)
+                    cert = sock.getpeercert(binary_form=True)
+                    assert cert
+                    got = self._hashfunc(cert).digest()
+                    expected = self._fingerprint
+                    if got != expected:
+                        raise FingerprintMismatch(expected, got, host, port)
+                return conn
+            except OSError as e:
+                exc = e
+        else:
+            raise ClientOSError('Can not connect to %s:%s' %
+                                (req.host, req.port)) from exc
 
 
 class ProxyConnector(TCPConnector):
@@ -368,12 +548,14 @@ class ProxyConnector(TCPConnector):
     Usage:
 
     >>> conn = ProxyConnector(proxy="http://some.proxy.com")
-    >>> resp = yield from request('GET', 'http://python.org', connector=conn)
+    >>> session = ClientSession(connector=conn)
+    >>> resp = yield from session.get('http://python.org')
 
     """
 
-    def __init__(self, proxy, *args, proxy_auth=None, **kwargs):
-        super().__init__(*args, **kwargs)
+    def __init__(self, proxy, *, proxy_auth=None, force_close=True,
+                 **kwargs):
+        super().__init__(force_close=force_close, **kwargs)
         self._proxy = proxy
         self._proxy_auth = proxy_auth
         assert proxy.startswith('http://'), (
@@ -386,8 +568,16 @@ class ProxyConnector(TCPConnector):
         """Proxy URL."""
         return self._proxy
 
+    @property
+    def proxy_auth(self):
+        """Proxy auth info.
+
+        Should be BasicAuth instance.
+        """
+        return self._proxy_auth
+
     @asyncio.coroutine
-    def _create_connection(self, req, **kwargs):
+    def _create_connection(self, req):
         proxy_req = ClientRequest(
             hdrs.METH_GET, self._proxy,
             headers={hdrs.HOST: req.host},
@@ -422,6 +612,7 @@ class ProxyConnector(TCPConnector):
             key = (req.host, req.port, req.ssl)
             conn = Connection(self, key, proxy_req,
                               transport, proto, self._loop)
+            self._acquired[key].append(conn._transport)
             proxy_resp = proxy_req.send(conn.writer, conn.reader)
             try:
                 resp = yield from proxy_resp.start(conn, True)
@@ -430,6 +621,7 @@ class ProxyConnector(TCPConnector):
                 conn.close()
                 raise
             else:
+                conn.detach()
                 if resp.status != 200:
                     raise HttpProxyError(code=resp.status, message=resp.reason)
                 rawsock = transport.get_extra_info('socket', default=None)
@@ -439,7 +631,7 @@ class ProxyConnector(TCPConnector):
                 transport.pause_reading()
                 transport, proto = yield from self._loop.create_connection(
                     self._factory, ssl=True, sock=rawsock,
-                    server_hostname=req.host, **kwargs)
+                    server_hostname=req.host)
 
         return transport, proto
 
@@ -454,12 +646,13 @@ class UnixConnector(BaseConnector):
     Usage:
 
     >>> conn = UnixConnector(path='/path/to/socket')
-    >>> resp = yield from request('GET', 'http://python.org', connector=conn)
+    >>> session = ClientSession(connector=conn)
+    >>> resp = yield from session.get('http://python.org')
 
     """
 
-    def __init__(self, path, *args, **kw):
-        super().__init__(*args, **kw)
+    def __init__(self, path, **kwargs):
+        super().__init__(**kwargs)
         self._path = path
 
     @property
@@ -468,6 +661,6 @@ class UnixConnector(BaseConnector):
         return self._path
 
     @asyncio.coroutine
-    def _create_connection(self, req, **kwargs):
+    def _create_connection(self, req):
         return (yield from self._loop.create_unix_connection(
-            self._factory, self._path, **kwargs))
+            self._factory, self._path))
