@@ -575,16 +575,17 @@ class BodyPartWriterTestCase(unittest.TestCase):
 
     def test_autoset_content_disposition(self):
         self.part.obj = open(__file__, 'rb')
+        self.addCleanup(self.part.obj.close)
         self.part._fill_headers_with_defaults()
         self.assertIn(CONTENT_DISPOSITION, self.part.headers)
         fname = os.path.basename(self.part.obj.name)
-        self.assertEquals(
+        self.assertEqual(
             'attachment; filename="{0}"; filename*=utf-8\'\'{0}'.format(fname),
             self.part.headers[CONTENT_DISPOSITION])
 
     def test_set_content_disposition(self):
         self.part.set_content_disposition('attachment', foo='bar')
-        self.assertEquals(
+        self.assertEqual(
             'attachment; foo=bar',
             self.part.headers[CONTENT_DISPOSITION])
 
