@@ -176,6 +176,8 @@ class TestWebFunctional(unittest.TestCase):
             resp = yield from request('POST', url, data=[f],
                                       loop=self.loop)
             self.assertEqual(200, resp.status)
+            resp.close()
+            f.close()
 
         self.loop.run_until_complete(go())
 
@@ -209,15 +211,18 @@ class TestWebFunctional(unittest.TestCase):
             resp = yield from request('POST', url, data=[f1, f2],
                                       loop=self.loop)
             self.assertEqual(200, resp.status)
+            resp.close()
 
         self.loop.run_until_complete(go())
+        f1.close()
+        f2.close()
 
     def test_release_post_data(self):
 
         @asyncio.coroutine
         def handler(request):
             yield from request.release()
-            chunk = yield from request.payload.readany()
+            chunk = yield from request.content.readany()
             self.assertIs(EOF_MARKER, chunk)
             return web.Response(body=b'OK')
 
@@ -227,6 +232,7 @@ class TestWebFunctional(unittest.TestCase):
             resp = yield from request('POST', url, data='post text',
                                       loop=self.loop)
             self.assertEqual(200, resp.status)
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -250,6 +256,7 @@ class TestWebFunctional(unittest.TestCase):
                 loop=self.loop)
 
             self.assertEqual(200, resp.status)
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -260,20 +267,24 @@ class TestWebFunctional(unittest.TestCase):
             app, _, url = yield from self.create_server(
                 'GET', '/static/' + filename
             )
-            app.router.add_static('/static', dirname)
+            app.router.add_static('/static', dirname, chunk_size=16)
 
             resp = yield from request('GET', url, loop=self.loop)
             self.assertEqual(200, resp.status)
             txt = yield from resp.text()
-            self.assertEqual('file content{}'.format(os.linesep), txt)
+            self.assertEqual('file content', txt.rstrip())
             ct = resp.headers['CONTENT-TYPE']
             self.assertEqual('application/octet-stream', ct)
             self.assertEqual(resp.headers.get('CONTENT-ENCODING'), None)
+            resp.close()
 
             resp = yield from request('GET', url + 'fake', loop=self.loop)
             self.assertEqual(404, resp.status)
+            resp.close()
+
             resp = yield from request('GET', url + '/../../', loop=self.loop)
             self.assertEqual(404, resp.status)
+            resp.close()
 
         here = os.path.dirname(__file__)
         filename = os.path.join(here, 'data.unknown_mime_type')
@@ -297,11 +308,15 @@ class TestWebFunctional(unittest.TestCase):
             ct = resp.headers['CONTENT-TYPE']
             self.assertEqual('image/jpeg', ct)
             self.assertEqual(resp.headers.get('CONTENT-ENCODING'), None)
+            resp.close()
 
             resp = yield from request('GET', url + 'fake', loop=self.loop)
             self.assertEqual(404, resp.status)
+            resp.close()
+
             resp = yield from request('GET', url + '/../../', loop=self.loop)
             self.assertEqual(404, resp.status)
+            resp.close()
 
         here = os.path.dirname(__file__)
         filename = os.path.join(here, 'software_development_in_picture.jpg')
@@ -324,6 +339,7 @@ class TestWebFunctional(unittest.TestCase):
             self.assertEqual('text/plain', ct)
             encoding = resp.headers['CONTENT-ENCODING']
             self.assertEqual('gzip', encoding)
+            resp.close()
 
         here = os.path.dirname(__file__)
         filename = os.path.join(here, 'hello.txt.gz')
@@ -381,6 +397,7 @@ class TestWebFunctional(unittest.TestCase):
                 loop=self.loop)
 
             self.assertEqual(200, resp.status)
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -420,6 +437,7 @@ class TestWebFunctional(unittest.TestCase):
 
             self.assertEqual(200, resp.status)
             self.assertTrue(expect_received)
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -459,6 +477,7 @@ class TestWebFunctional(unittest.TestCase):
                 loop=self.loop)
 
             self.assertEqual(200, resp.status)
+            resp.close(force=True)
 
             auth_err = True
             resp = yield from request(
@@ -466,6 +485,7 @@ class TestWebFunctional(unittest.TestCase):
                 expect100=True,  # wait until server returns 100 continue
                 loop=self.loop)
             self.assertEqual(403, resp.status)
+            resp.close(force=True)
 
         self.loop.run_until_complete(go())
 
@@ -490,6 +510,7 @@ class TestWebFunctional(unittest.TestCase):
                 loop=self.loop)
 
             self.assertEqual(404, resp.status)
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -514,6 +535,7 @@ class TestWebFunctional(unittest.TestCase):
                 loop=self.loop)
 
             self.assertEqual(405, resp.status)
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -530,6 +552,7 @@ class TestWebFunctional(unittest.TestCase):
             resp = yield from request('GET', url, loop=self.loop,
                                       version=HttpVersion10)
             self.assertEqual('close', resp.headers['CONNECTION'])
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -548,6 +571,7 @@ class TestWebFunctional(unittest.TestCase):
                                       headers=headers,
                                       version=HttpVersion(0, 9))
             self.assertEqual('close', resp.headers['CONNECTION'])
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -565,6 +589,7 @@ class TestWebFunctional(unittest.TestCase):
             resp = yield from request('GET', url, loop=self.loop,
                                       headers=headers, version=HttpVersion10)
             self.assertEqual('close', resp.headers['CONNECTION'])
+            resp.close()
 
         self.loop.run_until_complete(go())
 
@@ -582,5 +607,94 @@ class TestWebFunctional(unittest.TestCase):
             resp = yield from request('GET', url, loop=self.loop,
                                       headers=headers, version=HttpVersion10)
             self.assertEqual('keep-alive', resp.headers['CONNECTION'])
+            resp.close()
+
+        self.loop.run_until_complete(go())
+
+    def test_upload_file(self):
+
+        here = os.path.dirname(__file__)
+        fname = os.path.join(here, 'software_development_in_picture.jpg')
+        with open(fname, 'rb') as f:
+            data = f.read()
+
+        @asyncio.coroutine
+        def handler(request):
+            form = yield from request.post()
+            raw_data = form['file'].file.read()
+            self.assertEqual(data, raw_data)
+            return web.Response(body=b'OK')
+
+        @asyncio.coroutine
+        def go():
+            _, _, url = yield from self.create_server('POST', '/', handler)
+            resp = yield from request('POST', url,
+                                      data={'file': data},
+                                      loop=self.loop)
+            self.assertEqual(200, resp.status)
+            resp.close()
+
+        self.loop.run_until_complete(go())
+
+    def test_upload_file_object(self):
+
+        here = os.path.dirname(__file__)
+        fname = os.path.join(here, 'software_development_in_picture.jpg')
+        with open(fname, 'rb') as f:
+            data = f.read()
+
+        @asyncio.coroutine
+        def handler(request):
+            form = yield from request.post()
+            raw_data = form['file'].file.read()
+            self.assertEqual(data, raw_data)
+            return web.Response(body=b'OK')
+
+        @asyncio.coroutine
+        def go():
+            _, _, url = yield from self.create_server('POST', '/', handler)
+            f = open(fname, 'rb')
+            resp = yield from request('POST', url,
+                                      data={'file': f},
+                                      loop=self.loop)
+            self.assertEqual(200, resp.status)
+            resp.close()
+            f.close()
+
+        self.loop.run_until_complete(go())
+
+    def test_empty_content_for_query_without_body(self):
+
+        @asyncio.coroutine
+        def handler(request):
+            self.assertFalse(request.has_body)
+            return web.Response(body=b'OK')
+
+        @asyncio.coroutine
+        def go():
+            _, srv, url = yield from self.create_server('GET', '/', handler)
+            resp = yield from request('GET', url, loop=self.loop)
+            self.assertEqual(200, resp.status)
+            txt = yield from resp.text()
+            self.assertEqual('OK', txt)
+
+        self.loop.run_until_complete(go())
+
+    def test_empty_content_for_query_with_body(self):
+
+        @asyncio.coroutine
+        def handler(request):
+            self.assertTrue(request.has_body)
+            body = yield from request.read()
+            return web.Response(body=body)
+
+        @asyncio.coroutine
+        def go():
+            _, srv, url = yield from self.create_server('POST', '/', handler)
+            resp = yield from request('POST', url, data=b'data',
+                                      loop=self.loop)
+            self.assertEqual(200, resp.status)
+            txt = yield from resp.text()
+            self.assertEqual('data', txt)
 
         self.loop.run_until_complete(go())
